# -*- coding: utf-8 -*-
"""
Created on Mon May 22 21:09:12 2023

@author: atakan
"""

import os
import copy
# from time import time
from ctREFPROP.ctREFPROP import REFPROPFunctionLibrary
import numpy as np
# import CoolProp.CoolProp as CP
import carbatpy

VERBOSE = False
DEFAULTS = carbatpy.CB_DEFAULTS
TREND = DEFAULTS['Fluid_Defaults']['TREND']

if TREND["TREND_INSTALLED"]:
    trend_dll = TREND["TREND_DLL"]
    trend_path = TREND["TREND_PATH"]
    try:
        import fluid as tr_fl  # TREND fluids

    except ImportError as e:
        print(f"Import error for 'fluid': {e}")


# os.environ['RPPREFIX'] = r'C:/Program Files (x86)/REFPROP'
# os.environ['RPPREFIXs'] = r'C:/Program Files (x86)/REFPROP/secondCopyREFPROP'
_PROPS = DEFAULTS["Fluid_Defaults"]['PROPS']  # "REFPROP"  # or "CoolProp"

_fl_properties_names = DEFAULTS["Fluid_Defaults"]['Property_Names']
_THERMO_STRING = DEFAULTS["Fluid_Defaults"]['THERMO_STRING']
_THERMO_TREND = request =DEFAULTS["Fluid_Defaults"]['THERMO_TREND']  # careful density not volume
_TRANS_STRING = DEFAULTS["Fluid_Defaults"]['TRANS_STRING']
_TRANS_TREND = DEFAULTS["Fluid_Defaults"]['TRANS_TREND']


_T_SURROUNDING = DEFAULTS["General"]['T_SUR'] # K
_MODEL_ARGS = {}

# order for coolprop,alle_0:[_temp, p,  h, 1/ rho, s,x,cp, mu,  lambda_s,
# prandtl, phase]"
_UNITS = 21
rp_instance = ""

if _PROPS == "REFPROP":
    DLL_SELECT = DEFAULTS["Fluid_Defaults"]['DLL_SELECT']
    try:
        rp_instance = REFPROPFunctionLibrary(os.environ['RPPREFIX'])
        # be careful pressure is in Pa!
        _UNITS = rp_instance.GETENUMdll(0, DEFAULTS["Fluid_Defaults"]['UNITS']).iEnum
    except:  # ModuleNotFoundError as errefprop:
        print(f"Refprop is not installed!")  # " {errefprop}")
elif _PROPS == "TREND":
    pass
else:
    print(f"_PROPS value, not installed {_PROPS}")


class FluidModel:
    """
    Fluid model to be used Refprop or TREND

    Only model, chemical compunds units and for TREND some more values are set here. The
    mixture composition, state etc. is set in Fluid.
    """

    def __init__(self, fluid, units=_UNITS, props=_PROPS, rp_inst=rp_instance,
                 args=_MODEL_ARGS):
        """
        For Class Fluidmodell a fluid (mixture) must be defined, the evaluation
        takes place with props, units can be set and an instance can be set,
        the latter is important, if more than one fluid is used.

        Parameters
        ----------
        fluid : String
            as defined in the props (Model), for RefProp it is
            "fluid1 * fluid2 * fluid3".
        units : integer, optional
            units, normally SI wuth MASS base, see props(RefProp).
            The default is _UNITS.
        props : string, optional
            select the property model. The default is _PROPS.
        rp_inst : RefProp-Instance, optional
            where Refprop is installed, for two fluids, as in heat exchangers,
            two installations/instances are needed. The default is rp_instance.
        args : Dictionary, optional
            Further arguments passed to the module used for fluid property
            evaluation. for props =="REFPROP" nothing is needed. The default is
            the empty dictionary _MODEL_ARGS.

        Returns
        -------
        None.

        """
        self.fluid = fluid
        self.props = props
        self.units = units
        self.args = args
        if props == "REFPROP":
            self.rp_instance = rp_inst
            self.set_rp_fluid()
        elif props == "TREND":
            self.set_tr_fluid()

    def set_rp_fluid(self, modwf=REFPROPFunctionLibrary, name='RPPREFIX'):
        """
        A new instance of Refpropdll for the given fluid. It can then be called
        using fluid =""

        Parameters
        ----------
        fluid : string
            fluid (mixture) name, as described in REFPROP.

        Returns
        -------
        self.rp_instance : REFPROP Instance
            for further usage.

        """

        self.rp_instance = modwf(os.environ[name])
        self.rp_instance.SETPATHdll(os.environ[name])
        ierr = self.rp_instance.SETFLUIDSdll(self.fluid)
        if ierr != 0:
            print(f"Fehler in setfluid {ierr}")
            print(self.rp_instance.ERRMSGdll(ierr))
        return self.rp_instance

    def set_tr_fluid(self):
        self.fluid = self.fluid_to_list()
        if self.units == 21:
            self.units = "specific"

        n_compounds = len(self.fluid)
        if len(self.args["moles"]) < n_compounds:
            self.args["moles"] = list(np.zeros(n_compounds))
            self.args["moles"][0] = 1.
        if len(self.args["eos_ind"]) < n_compounds:
            self.args["eos_ind"] = list(np.ones(n_compounds))
        if self.args["mix_ind"] < 1:
            self.args["mix_ind"] = 1

        _trend_dict = {"Input": "TP",
                       'calctype': "H",
                       'fluids': self.fluid,
                       "moles": self.args["moles"],
                       "eos_ind": self.args["eos_ind"],
                       'mix_ind': self.args["mix_ind"],
                       'path': trend_path,
                       'unit': self.units,
                       'dll_path': trend_dll}

        self.rp_instance = tr_fl.Fluid(*_trend_dict.values())
        err_flag = self.rp_instance.errorflag.value
        if err_flag > 0:
            print(f"fluid-Trend problem, errorflag:{err_flag}")
        return self.rp_instance

    def fluid_to_list(self):
        """
        conversts the fluid names string from Refprop to list for TREND

        Returns
        -------
        List of strings
            the fluid names in a list.

        """
        no_blank = self.fluid.replace(" ", "")
        return no_blank.split("*")


class FluidState:
    """
    The thermodynamic state of a fluid is stored here
    """

    def __init__(self, state, what, **kwargs):
        self.props = kwargs.get("props", _PROPS)
        self. verbose = kwargs.get("verbose", False)
        self.liq_x = kwargs.get("x", None)
        self.vap_y = kwargs.get("y", None)
        self.total_z = kwargs.get("z", None)

        self.no_val = what.count(";")+1
        self.temperature = state[0]
        self.pressure = state[1]
        self.sp_volume = state[3]
        self.enthalpy = state[2]
        self.entropy = state[4]
        self.quality = state[5]
        self.int_energy = state[6]
        self.state = state[:7]
        self.prop_names = _fl_properties_names
        if self.verbose:
            print(what, no_val, state)

        if self.no_val > 7 and self.props in ("REFPROP",):
            # ";VIS;TCX;PRANDTL;KV;M;W"
            self.viscosity = state[7]
            self.thermal_conductivity = state[8]
            self.prandtl = state[9]
            self.kin_viscosity = state[10]
            self.molecular_mass = state[11]
            self.speed_of_sound = state[12]
            self.transport = state[7:13]
            self.state = state[:13]

    def state_to_dict(self):
        return_dict = dict(zip(_fl_properties_names[:self.no_val], self.state[:self.no_val]))
        return_dict["x"] = self.liq_x
        return_dict["y"] = self.vap_y
        return_dict["z"] = self.total_z
        return  return_dict


class Fluid:
    """
    The Fluid class is used to set, get, and print states of  a fluid with a given

    model (e.g. RefProp). The compounds are set in the fluidmodel, while the
    composition is also set here.

    """

    def __init__(self, fluidmodel, composition=None, option=1):
        """
        Generate a Fluid instance

        a FluidModel instance, the composition and an unused option has to be passed.

        Parameters
        ----------
        fluidmodel : FluidModel
            the instance of the property model to be used.
        composition : list, optional
            mole fractions of all compounds. The default is [1.0].
        option : TYPE, optional
            unused. The default is 1.

        Returns
        -------
        None.

        """

        self.fluidmodel = fluidmodel
        if composition is None:
            composition = [1.0]
        else:
            self.composition = composition
        self.composition = copy.copy(composition)
        self.properties = None
        self.state_v = None
        self.option = option
        self.no_compounds = len(composition)
        self.comp_mass = np.zeros(self.no_compounds)
        self.herr = 0
        self.props = fluidmodel.props
        # BA something is wrong here (probably 2024-07-10
        self.set_composition(composition)
        if self.props == "TREND":
            # important for trend mass->mole"

            self.fluidmodel.rp_instance.set_moles(self.composition)
            if VERBOSE:
                print("Konversion", self.composition,
                      self.comp_mass,
                      self.composition,
                      self.fluidmodel.rp_instance.get_moles())

    def set_composition(self, composition, **kwargs):
        """
        Mainly for TREND mole fractions have to converted to mass fractions

        Sets the self.composition used in the Fluid settings and the self.comp_mass
        and self.composition by calling calc_mass_fraction()

        Parameters
        ----------
        composition : list/numpy.array
            mole fractions must sum up to 1.

        Returns
        -------
        None.

        """
        verbose = kwargs.get("verbose", False)
        if verbose:
            print("set_comp", composition, self.composition, self.comp_mass)
        self.composition = composition
        if self.fluidmodel.props == "TREND":
            self.comp_mass = composition
            self.calc_mass_fraction()
            self.fluidmodel.rp_instance.set_moles(self.composition)
            # here is the mass fraction to internal mol fraction conversion
        elif self.fluidmodel.props == "REFPROP":
            self.comp_mass = self._rp_mass_from_mole(composition)
        if verbose:
            print("set_comp-after", composition, self.composition, self.comp_mass)

    def calc_mass_fraction(self):
        """ mole to mass fractions for TREND"""
        mass_frac = np.zeros(self.no_compounds)
        for n_c in range(self.no_compounds):
            mass_frac[n_c] = self.fluidmodel.rp_instance.get_mw(n_c) * \
                self.composition[n_c]
        self.comp_mass = mass_frac/mass_frac.sum()
        self.composition = copy.copy(self.comp_mass)
        return self.composition

    def set_state(self, values, given="TP", wanted=_THERMO_STRING, composition=None,
                  **kwargs):
        """
        Sets the state of a fluid and calculates the wanted properties.

        Parameters
        ----------
        values : list[2]
            the two state parameter values.
        given : String, optional
            What are the two values (T,P,H,S,Q,V). The default is "TP".
        wanted : String, optional
            Which parameters shall be calculated? The default is _THERMO_STRING.
        composition : list, optional
            mole fraction of each compound, must sum up to 1. When empty, the last/actual
            value  will be used. The default is [].
        kwargs : dictionary, optional
            Flags like iMass or iFlag can be passed to Refprop. with iMass=1 the input
            composition is taken as mass fraction, otherwise they are mole fractions.

        Raises
        ------
        Exception
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.
        """

        composition = self._update_composition(composition)

        if self.fluidmodel.props == "REFPROP":
            self._set_state_refprop(values, given, wanted, composition, **kwargs)
        elif self.fluidmodel.props == "TREND":
            self._set_state_trend(values, given, wanted)
        else:
            raise NotImplementedError(
                f"Property model {self.fluidmodel.props} not implemented yet!")

        return np.array([*self.properties.state])

    def _update_composition(self, composition):
        if composition is None:  # mole fractions only!
            return self.composition

        self.set_composition(composition)  # important for trend mass->mole
        if VERBOSE:
            print("comp-set", composition, self.composition,
                  self.comp_mass, self.composition)
        return composition

    def _set_state_refprop(self, values, given, wanted, composition, **kwargs):
        i_mass = kwargs.get("iMass", 0)
        i_flag = kwargs.get("iFlag", 0)  # if this is 1 most tests fail!
        dll_select = kwargs.get("dll_select", "2dll")
        verbose = kwargs.get("verbose", False)

        match dll_select:  # noqa
            case "2dll":
                i_mass=0
                state = self.fluidmodel.rp_instance.REFPROP2dll(
                    self.fluidmodel.fluid, given, wanted,
                    self.fluidmodel.units,
                    i_flag, values[0], values[1],
                    copy.copy(self.composition)
                )
            case "dll":
                i_mass=1
                state = self.fluidmodel.rp_instance.REFPROPdll(
                    self.fluidmodel.fluid, given, wanted,
                    self.fluidmodel.units,
                    i_mass,
                    i_flag, values[0], values[1],
                    copy.copy(self.composition)
                )
                
                if verbose:
                    print(state.z[:5], self.composition, self.comp_mass)
            case _:
                raise ValueError(f"This iMass value is not allowed {i_mass}")

        if state.ierr == 0:
            state_val = state.Output
            z_act = state.z[:self.no_compounds]
            x_act = y_act = None

            if state.q > 0 and state.q < 1:  # mole -> mass quality problem of Refprop
                # state_val[5] = state.q
                if dll_select == "dll":
                    x_act = state.x[:self.no_compounds]
                    y_act = state.y[:self.no_compounds]
                    if _UNITS == 21:  # convert back to mole fractions
                        x_act = self._rp_mole_from_mass(state.x)
                        y_act = self._rp_mole_from_mass(state.y)
                        z_act = self._rp_mole_from_mass(state.z)
            self.properties = FluidState(state_val, wanted, x=x_act, y=y_act, z=z_act)
            self.val_dict = self.properties.state_to_dict()

        else:
            self.herr = state.herr
            raise RuntimeError(f"Property-Refprop problem: {state.herr}!")

    def _rp_mass_from_mole(self, composition):
        return self.fluidmodel.rp_instance.XMASSdll(composition).xkg[:self.no_compounds]

    def _rp_mole_from_mass(self, composition):
        return self.fluidmodel.rp_instance.XMOLEdll(composition).xmol[:self.no_compounds]

    def _set_state_trend(self, values, given, wanted):
        if VERBOSE:
            print("in TREND set state")

        values, given = self._adjust_trend_values(values, given)
        all_values = self.fluidmodel.rp_instance.ALLPROP(given, values[0], values[1])
        all_val_initial = copy.copy(all_values)

        if all_values["T"] > 0:  # on error it is -8888
            all_values["P"] = all_values["P"] * 1e6  # MPa
            all_values["D"] = 1 / all_values["D"]  # V (spec. volume)
            all_values = [all_values[key] for key in _THERMO_TREND]
            if wanted == _TRANS_STRING:

                trans_dict = {}
                speed_sound = all_val_initial["WS"]
                for tr_want in _TRANS_TREND[-3:]:
                    self.fluidmodel.rp_instance.set_calctype(tr_want)
                    self.fluidmodel.rp_instance.set_input(given)

                    trans_dict[tr_want], err = self.fluidmodel.rp_instance.TREND_EOS(
                        *values)
                    print(tr_want, ":", trans_dict[tr_want], err)
                    if trans_dict[tr_want] < 0:
                        print("WARNING: Problem / NOT IMPLEMENTED IN TREND!")

            self.properties = FluidState(all_values, wanted, props="TREND")
            self.val_dict = self.properties.state_to_dict()
        else:
            raise RuntimeError(
                f"property calculation in TREND failed, error code TREND: {all_values['T']}")

    def _adjust_trend_values(self, values, given):
        pressure_in = given[:2].find("P")
        quality_in = given[:2].find("Q")
        temp_in = given[:2].find("T")

        if pressure_in >= 0:
            values[pressure_in] = values[pressure_in] / 1e6  # MPa in Trend
        if quality_in > 0:
            if pressure_in >= 0:
                values, given = self._adjust_values_for_quality_pressure(
                    values, quality_in, pressure_in)
            elif temp_in >= 0:
                values, given = self._adjust_values_for_quality_temperature(
                    values, quality_in, temp_in)
            else:
                raise NotImplementedError(
                    f'Error: Qualities cannot be combined with {given}!')
        return values, given

    def _adjust_values_for_quality_pressure(self, values, quality_in, pressure_in):
        match values[quality_in]:
            case 0.:
                given = "PLIQ"
            case 1.:
                given = "PVAP"
            case _:
                raise ValueError('Error: only qualities 0 or 1 for mixtures!')
        values[0] = values[pressure_in]
        values[1] = 10
        return values, given

    def _adjust_values_for_quality_temperature(self, values, quality_in, temp_in):
        match values[quality_in]:
            case 0.:
                given = "TLIQ"
            case 1.:
                given = "TVAP"
            case _:
                raise ValueError('Error: only qualities 0 or 1 for mixtures!')
        values[0] = values[temp_in]
        values[1] = 10
        return values, given

    def set_state_v(self, values, given="TP", wanted=_THERMO_STRING, **kwargs):
        """
        set many Fluid states with the values given as array(vector)

        Parameters
        ----------
        values : numpy.array((N,2))
            the N value pairs to fix the states.
        given : String, optional
            what are the input values?(T,P, Q, S,...). The default is "TP".
        wanted : String, optional
            with all the outputs wanted. The default is _THERMO_STRING.

        Returns
        -------
        output : numpy.array(N, n_wanted)
            all the wanted properties for the N inputs.

        """
        dimension = np.shape(values)
        number_wanted = wanted.count(";")+1
        output = np.zeros((dimension[0], number_wanted))
        for count, value in enumerate(values):
            output[count, :] = self.set_state(value, given, wanted, **kwargs)
        self.state_v = output
        return output

    def print_state(self):
        """ Print the actual state of the Fluid"""
        pr = self.properties
        flmod = self.fluidmodel
        print(f"\n{flmod.fluid}, composition: {self.composition}")
        print(f"T:{pr.temperature:.2f} K, p: {pr.pressure/1e5 :.2f} bar,  h: {pr.enthalpy/1000: .2f} kJ/kg, s: {pr.entropy/1000:.3f} kJ/kg K\n")
        if pr.quality >= 0:
            print(f"Quality: {pr.quality :.3f}")

    def calc_temp_mean(self, h_final):
        """
        Calculate the thermodynamic mean temperature between the actual state

        and the final enthalpy along an isobaric (Delta h /Delta s)

        Parameters
        ----------
        h_final : float
            enthalpy of he final stat.

        Returns
        -------
        temp_mean : float
            the thermodynamic mean temperature.

        """
        actual_props = self.properties.state
        final_props = self.set_state([h_final, actual_props[1]], "HP")
        temp_mean = (final_props[2] - actual_props[2]) \
            / (final_props[4] - actual_props[4])
        return temp_mean


def init_fluid(fluid, composition, **keywords):
    """
    short way to define a Fluid and a FluidModel

    Parameters
    ----------
    fluid : string
        The species within the fluid.
    composition : List
        mole fraction for each fluid.
    **keywords : TYPE
        all keywords needed for the FluidModel, if non-defults shall be set.

    Returns
    -------
    actual_fluid : Fluid
        Instance of the actually set Fluid.

    """
    flmod = FluidModel(fluid, **keywords)
    actual_fluid = Fluid(flmod, composition)
    return actual_fluid


if __name__ == "__main__":
    FLUID = "Propane * Pentane"
    comp = [.50, 0.5]
    P_ACT = 1e5
    T_ACT = 370
    flm = FluidModel(FLUID)
    myFluid = Fluid(flm, comp)
    st00 = myFluid.set_state([T_ACT, P_ACT], "TP")
    st1 = myFluid.set_state([T_ACT, P_ACT], "TP",
                            _TRANS_STRING)
    print(st00, st1)
    myFluid.print_state()
    myFluid.set_composition([.2, .8])
    st0 = myFluid.set_state([T_ACT, P_ACT], "TP", composition=[.35, .65])
    myFluid.print_state()

    mean_temp_act = myFluid.calc_temp_mean(st0[2]+1e5)
    print(f"Mean Temperature {mean_temp_act} K")

    # value_vec = np.array([[300, 1e5], [400, 1e5], [500, 1e5]])
    # stv = myFluid.set_state_v(value_vec, "TP")

    # print(myFluid.set_state_v(value_vec, "TP"))
    # print(myFluid.set_state([300., 1.], "TQ"))

    # New simple way to get an instance of Fluid
    myFluid2 = init_fluid(FLUID, comp)

    # Refprop with mass fractions as input
    print('\n---------------- Refprop Mass fraction Input')
    FLUID = "Propane * Pentane"
    comp = [.5, .5]
    myFluid = init_fluid(FLUID, comp)

    quality = 0.5       # is interpreted by carbatpy as mol/mol to calculate state
    # but given back in string as kg/kg
    st0 = myFluid.set_state([300., quality], "TQ", dll_select="dll")
    myFluid.print_state()
    st0n = myFluid.set_state([st0[1], st0[2]], "PH", dll_select="dll")
    myFluid.print_state()

    print(f"quality prescribed: {quality}")
    print(f"resulting quality from carbatpy in st0: {st0[5]}")

    print("\n------------\n")

    # Trend:
    my_dict = {"Input": "TP",
               'calctype': "H",
               'fluids': FLUID,
               "moles": comp,
               "eos_ind": [1, 1],
               'mix_ind': 1,
               'path': trend_path,
               'unit': 'specific',
               'dll_path': trend_dll}
    # comp =[.5,0.5]

    flm_trend = FluidModel(FLUID, props="TREND", args=my_dict)
    # [0.6206645653468038, 0.3793354346531962])
    my_tr_fl = Fluid(flm_trend, comp)
    state_trend = my_tr_fl.set_state([T_ACT, P_ACT], "TP")
    print("\nComparison Refprop vs. Trend\n", st00, "\n", state_trend,
          my_tr_fl.properties.enthalpy)
    state_tr_trans = my_tr_fl.set_state([T_ACT, P_ACT], "TP", wanted=_TRANS_STRING)
    print(f"Trans-String: {state_tr_trans}")

    # shortcut to generate a trend fluid
    flt0 = carbatpy.init_fluid(FLUID, comp, props="TREND", args=my_dict)

    # saturated mixture:
    state_trend_satp = my_tr_fl.set_state([P_ACT, 0], "PQ")
    print("\nSaturated mixture, liq:\n", state_trend_satp,  my_tr_fl.properties.enthalpy)
    state_trend_satp = my_tr_fl.set_state([P_ACT, 1], "PQ")
    print("Saturated mixture, vap:\n", state_trend_satp,  my_tr_fl.properties.enthalpy)

    # some temperature dependence
    state_trend_c = my_tr_fl.set_state([230., P_ACT], "TP")

    enth = np.linspace(state_trend_c[2], state_trend[2], 29)
    all_props_act = []
    for h_act in enth:

        all_props_act.append(my_tr_fl.set_state([P_ACT, h_act], "PH"))
    all_props_act = np.array(all_props_act)

    import matplotlib.pyplot as plt
    fi, ax = plt.subplots(1)
    ax.plot(all_props_act[:, 2], all_props_act[:, 0])



# -*- coding: utf-8 -*-
"""
Created on Tue Jul 16 21:25:43 2024

@author: atakan
"""

import copy
import numpy as np
from scipy.optimize import minimize
import matplotlib.pyplot as plt
import carbatpy as cb

class StaticHeatExchanger:
    """
    Static counter-flow heat exchanger class.
    """

    def __init__(self, fluids, h_dot_min, h_out_w, h_limit_s=np.nan, points=50, 
                 d_temp_separation_min=0.5, calc_type="const", name="evaporator"):
        """
        Initialize the heat exchanger with given parameters.

        Parameters
        ----------
        fluids : list of 2 fprop.Fluid
            Two fluids entering the heat exchanger, typically at room temperature.
        h_dot_min : float
            Enthalpy flow rate (W) to be transferred.
        h_out_w : float
            Exit enthalpy of the working fluid.
        h_limit_s : float or nan, optional
            Limit in enthalpy for the secondary fluid. Default is nan.
        points : int, optional
            Number of points for checking minimum approach temperature. Default is 50.
        d_temp_separation_min : float, optional
            Minimum approach temperature between the two fluids. Default is 0.5.
        calc_type : str, optional
            Calculation type. Default is "const".
        name : str, optional
            Name of the heat exchanger. Default is "evaporator".
        """
        self.fluids = fluids
        state_in_w = fluids[0].properties.state
        self.m_dot_s = 0
        self.h_limit_s = h_limit_s
        self.q_dot = h_dot_min
        self.m_dot_w = np.abs(h_dot_min / (h_out_w - state_in_w[2]))
        self.h_out_w = h_out_w
        self.h_out_s = float(copy.copy(h_limit_s))
        self.points = points
        self.d_temp_separation_min = d_temp_separation_min
        self.heating = 1 if h_out_w < state_in_w[2] else -1
        self.calc_type = calc_type
        self.name = name
        self.all_states = np.zeros((points, len(cb.fprop._THERMO_STRING.split(";"))))
        self.h_in_out = np.zeros((2, 4))
        self.dt_mean = None
        self.dt_min = None
        self.dt_max = None
        self.warning = 0
        self.warning_message = "All o.k."

    def pinch_calc(self, verbose=False, plot_info={}):
        """
        Calculate the enthalpy and temperature changes in the heat exchanger.

        Parameters
        ----------
        verbose : bool, optional
            If True, print several variables. Default is False.
        plot_info : dict, optional
            Dictionary with plotting information. Default is empty.

        Returns
        -------
        m_dot_s : float
            Secondary fluid mass flow rate (kg/s).
        d_tempall : numpy.ndarray
            Temperature differences along the counter-flow heat exchanger.
        w_array : numpy.ndarray
            Properties of the working fluid along the heat exchanger.
        s_array : numpy.ndarray
            Properties of the secondary fluid along the heat exchanger.
        """
        self.warning = 0
        self.warning_message = "All o.k."
        w_in, s_in = copy.copy(self.fluids[0]), copy.copy(self.fluids[1])
        w_out, s_out = copy.copy(self.fluids[0]), copy.copy(self.fluids[1])

        self.h_in_out[1, 0] = s_in.properties.enthalpy
        self.h_in_out[0, 0] = w_in.properties.enthalpy
        self.h_in_out[0, 1] = self.h_out_w

        state_out_s = s_out.set_state([self.h_out_s, s_in.properties.pressure], "HP")
        self.h_in_out[1, 1] = state_out_s[2]
        h_delta_s = np.abs(state_out_s[2] - s_in.properties.enthalpy)
        self.m_dot_s = self.q_dot / h_delta_s

        s_array = self._calculate_state_array(s_out, self.h_in_out[1, :2], self.points)
        w_array = self._calculate_state_array(w_out, self.h_in_out[0, 1::-1], self.points)

        d_tempall = w_array[:, 0] - s_array[:, 0]
        self.dt_mean, self.dt_min, self.dt_max = d_tempall.mean(), np.abs(d_tempall).min(), np.abs(d_tempall).max()

        self._check_temperature_consistency(d_tempall)

        if plot_info:
            self._plot_heat_exchanger(plot_info, w_array, s_array)

        if verbose:
            self._print_verbose(d_tempall)

        return self.m_dot_s, d_tempall * self.heating, w_array, s_array

    def pinch_root(self, h_out_s, secondary, verbose=False):
        """
        Find the minimum mean temperature difference in the heat exchanger.

        Parameters
        ----------
        h_out_s : float
            The output enthalpy, default for the secondary fluid.
        secondary : bool
            Whether to vary the secondary fluid output (True) or the working fluid (False).

        Returns
        -------
        float
            The mean temperature difference.
        """
        value = h_out_s if isinstance(h_out_s, float) else h_out_s[0]
        self.h_out_s = value if secondary else self.h_out_w
        self.h_out_w = value if not secondary else self.h_out_s

        mdot_s, d_temps, wf_states, sf_states = self.pinch_calc()

        return abs(self.dt_mean) if self.warning < 100 else 500.0

    def find_pinch(self, secondary=True):
        """
        Vary the secondary fluid enthalpy until the minimum approach temperature is reached.

        Parameters
        ----------
        secondary : bool, optional
            Whether to vary the secondary fluid (True) or the working fluid (False). Default is True.

        Returns
        -------
        float
            The optimized enthalpy of the secondary fluid.
        """
        x0 = copy.copy(self.h_out_s if secondary else self.h_out_w)
        tolerance = 1e-3

        try:
            result = minimize(self.pinch_root, x0, args=(secondary,), method='Nelder-Mead', tol=tolerance)

            if result.success or result.status == 2:
                if result.status == 2:
                    self.warning = 2
                    self.warning_message = "Minimization problem: " + result.message
                self.h_out_s = result.x[0] if secondary else self.h_out_w
                self.h_out_w = result.x[0] if not secondary else self.h_out_s
                return result.x[0]

        except Exception as e:
            self._handle_exception(e)
            return 10000

        self.warning = 1
        return self.warning

    def pinch_plot(self, plot_fname="", plotting=True):
        """
        Calculate and plot the minimum approach temperature of the heat exchanger.

        Parameters
        ----------
        plot_fname : str, optional
            Filename to save the plot. Default is "".
        plotting : bool, optional
            Whether to plot the results. Default is True.

        Returns
        -------
        m_dot_s : float
            Mass flow rate of the secondary fluid in kg/s.
        d_tempall : np.ndarray
            Temperature differences between the two fluids along the heat exchanger.
        w_array : np.ndarray
            States of the working fluid along the heat exchanger.
        s_array : np.ndarray
            States of the secondary fluid along the heat exchanger.
        """
        m_dot_s, d_tempall, w_array, s_array = self.pinch_calc()

        if plotting:
            self._plot_results(w_array, s_array, plot_fname)

        return m_dot_s, d_tempall, w_array, s_array

    def _calculate_state_array(self, fluid, h_range, points):
        h_array = np.linspace(h_range[0], h_range[1], points)
        values = np.zeros((points, 2))
        values[:, 0] = h_array
        values[:, 1] = fluid.properties.pressure
        return fluid.set_state_v(values, "HP")

    def _check_temperature_consistency(self, d_tempall):
        if np.abs(self.dt_min) - self.d_temp_separation_min < 1e-3:
            self.warning = 900
            self.warning_message = "Below minimum approach temperature!"

        if np.any(d_tempall > 0) and np.any(d_tempall < 0):
            self.warning = 999
            self.dt_mean = 1e6
            self.warning_message = "Temperatures crossing or wrong side!"

    def _plot_heat_exchanger(self, plot_info, w_array, s_array):
        if plot_info["what"][0] == 2:
            data_w = (w_array[:, plot_info["what"][0]] - w_array[:, plot_info["what"][0]].min()) * self.m_dot_w + plot_info["x-shift"][0]
            data_s = (s_array[:, plot_info["what"][0]] - s_array[:, plot_info["what"][0]].min()) * self.m_dot_s + plot_info["x-shift"][1]
            plot_info["ax"].plot(data_w, w_array[:, plot_info["what"][1]], plot_info["col"][0], label=plot_info["label"][0])
            plot_info["ax"].plot(data_s, s_array[:, plot_info["what"][1]], plot_info["col"][1], label=plot_info["label"][1])
        else:
            print(f"H-Ex: plotting only implemented for T-H_dot [2,0]. You requested {plot_info['what']}")

    def _print_verbose(self, d_tempall):
        print(f"Min T-distance {self.dt_min}, Mean T-distance {self.dt_mean}")
        if self.heating > 0:
            print("cond", d_tempall[0], d_tempall[-1], d_tempall.min(), d_tempall.max())
        else:
            print("evap", d_tempall[0], d_tempall[-1], d_tempall.max(), d_tempall.min())

    def _handle_exception(self, e):
        print("find pinch:", type(e))
        print(e.args)
        print(e)
        print("root-exception", self.heating)

    def _plot_results(self, w_array, s_array, plot_fname):
        h_w_plot_array = (w_array[:, 2] - w_array[:, 2].min()) * self.m_dot_w
        fig, ax_one = plt.subplots(1, 1)
        ax_one.plot((s_array[:, 2] - s_array[:, 2].min()) * self.m_dot_s, s_array[:, 0], "v")
        ax_one.plot(h_w_plot_array, w_array[:, 0], "o")
        ax_one.set_xlabel("specific enthalpy flow per mass of secondary fluid / (J / kg)")
        ax_one.set_ylabel("temperature / (K)")
        ax_one.set_title("heat exchanger, simple")
        if plot_fname:
            fig.savefig(plot_fname)



# -*- coding: utf-8 -*-
"""
heat_exchanger_thermo_vs von ChatGPT überarbeitet
Created on Tue Jul 16 21:30:53 2024

@author: atakan
"""

import copy
import numpy as np
from scipy.optimize import minimize
import matplotlib.pyplot as plt
import carbatpy as cb

class StaticHeatExchanger:
    """
    Class for static counter-flow heat exchanger.

   This class models a static counter-flow heat exchanger with no time dependence and no use of UA values (heat transfer coefficients * areas).
   Instead, it tries to meet a minimum approach temperature by varying the mass flow rates and possibly the working fluid pressure in the future.
   Both the first law and the second law of thermodynamics are checked.

   Parameters
   ----------
   fluids : list of fprop.Fluid
       The definition of the two fluids as they enter the heat exchanger, typically at room temperature.
   h_dot_min : float
       The enthalpy flow rate (W) that needs to be transferred.
   h_out_w : float
       The exit enthalpy of the working fluid.
   h_limit_s : float or nan, optional
       If there is a limit in enthalpy for the secondary fluid, it can be given here. The default is nan.
   points : int, optional
       The number of points (array) at which the minimum approach temperature will be checked and properties returned. The default is 30.
   d_temp_separation_min : float, optional
       Minimum approach temperature (pinch point) between the two fluids. The default is 0.5.
   calc_type : str, optional
       The type of calculation to perform; only one type is implemented so far. The default is "const".
   name : str, optional
       Name of the heat exchanger. The default is "evaporator".

   Attributes
   ----------
   all_states : np.ndarray
       Array to store the state properties of the fluids along the heat exchanger.
   dt_mean : float
       Mean temperature difference between the fluids.
   dt_min : float
       Minimum temperature difference between the fluids.
   dt_max : float
       Maximum temperature difference between the fluids.
   warning : int
       Warning code indicating the status of the calculation.
   warning_message : str
       Message describing any warnings that occurred during the calculation.

   Notes
   -----
   The minimum approach temperature is crucial for the design and operation of heat exchangers as it impacts both efficiency and feasibility.
   Future improvements will include varying the working fluid pressure to achieve the desired temperature approach.

   Examples
   --------
   >>> hex = StaticHeatExchanger([fluid1, fluid2], 1000, 200)
   >>> hex.pinch_calc(verbose=True)
    """

    def __init__(self, fluids, h_dot_min, h_out_w, **kwargs):
        """
        Initialize the heat exchanger with given parameters.

        Parameters
        ----------
        fluids : list of 2 fprop.Fluid
            Two fluids entering the heat exchanger, typically at room temperature.
        h_dot_min : float
            Enthalpy flow rate (W) to be transferred.
        h_out_w : float
            Exit enthalpy of the working fluid.
        kwargs : dict, optional
            Optional parameters for the heat exchanger configuration:
                - h_limit_s : float or nan, default np.nan
                - points : int, default 50
                - d_temp_separation_min : float, default 0.5
                - calc_type : str, default "const"
                - name : str, default "evaporator"
                - plot_info : dict, default {}
        """
        self.fluids = fluids
        state_in_w = fluids[0].properties.state

        self.h_dot_min = h_dot_min
        self.h_out_w = h_out_w

        self.h_limit_s = kwargs.get('h_limit_s', np.nan)
        self.points = kwargs.get('points', 50)
        self.d_temp_separation_min = kwargs.get('d_temp_separation_min', 0.5)
        self.calc_type = kwargs.get('calc_type', "const")
        self.name = kwargs.get('name', "evaporator")
        self.plot_info = kwargs.get('plot_info', {})

        self.m_dot_s = 0
        self.q_dot = h_dot_min
        self.m_dot_w = np.abs(h_dot_min / (h_out_w - state_in_w[2]))
        self.h_out_s = float(copy.copy(self.h_limit_s))
        self.heating = 1 if h_out_w < state_in_w[2] else -1

        self.all_states = np.zeros((self.points, len(cb.fprop._THERMO_STRING.split(";"))))
        self.h_in_out = np.zeros((2, 4))
        self.dt_mean = None
        self.dt_min = None
        self.dt_max = None
        self.warning = 0
        self.warning_message = "All o.k."

    @property
    def plot_info(self):
        return self._plot_info

    @plot_info.setter
    def plot_info(self, value):
        if not isinstance(value, dict):
            raise ValueError("plot_info must be a dictionary")
        self._plot_info = value

    def pinch_calc(self, verbose=False):
        """
        Calculate the enthalpy and temperature changes in the heat exchanger.
        Parameters
        ----------
        verbose : bool, optional
            If True, print several variables. Default is False.
        Returns
        -------
        m_dot_s : float
            Secondary fluid mass flow rate (kg/s).
        d_tempall : numpy.ndarray
            Temperature differences along the counter-flow heat exchanger.
        w_array : numpy.ndarray
            Properties of the working fluid along the heat exchanger.
        s_array : numpy.ndarray
            Properties of the secondary fluid along the heat exchanger.
        """
        self.warning = 0
        self.warning_message = "All o.k."
        w_in, s_in = copy.copy(self.fluids[0]), copy.copy(self.fluids[1])
        w_out, s_out = copy.copy(self.fluids[0]), copy.copy(self.fluids[1])
    
        self.h_in_out[1, 0] = s_in.properties.enthalpy
        self.h_in_out[0, 0] = w_in.properties.enthalpy
        self.h_in_out[0, 1] = self.h_out_w
    
        state_out_s = s_out.set_state([self.h_out_s, s_in.properties.pressure], "HP")
        self.h_in_out[1, 1] = state_out_s[2]
        h_delta_s = np.abs(state_out_s[2] - s_in.properties.enthalpy)
        self.m_dot_s = self.q_dot / h_delta_s
    
        s_array = self._calculate_state_array(s_out, self.h_in_out[1, :2], self.points)
        w_array = self._calculate_state_array(w_out, self.h_in_out[0, 1::-1], self.points)
    
        d_tempall = w_array[:, 0] - s_array[:, 0]
        self.dt_mean, self.dt_min, self.dt_max = d_tempall.mean(), np.abs(d_tempall).min(), np.abs(d_tempall).max()
    
        # Debugging-Ausgabe für Temperaturdifferenzen
        print(f"d_tempall: {d_tempall}")
        print(f"Mean T-distance: {self.dt_mean}, Min T-distance: {self.dt_min}, Max T-distance: {self.dt_max}")
    
        self._check_temperature_consistency(d_tempall)
    
        if self.plot_info:
            self._plot_heat_exchanger(w_array, s_array)
    
        if verbose:
            self._print_verbose(d_tempall)
    
        return self.m_dot_s, d_tempall * self.heating, w_array, s_array

    def pinch_root(self, h_out_s, secondary, verbose=False):
        """
        Find the minimum mean temperature difference in the heat exchanger.

        Parameters
        ----------
        h_out_s : float
            The output enthalpy, default for the secondary fluid.
        secondary : bool
            Whether to vary the secondary fluid output (True) or the working fluid (False).

        Returns
        -------
        float
            The mean temperature difference.
        """
        value = h_out_s if isinstance(h_out_s, float) else h_out_s[0]
        self.h_out_s = value if secondary else self.h_out_w
        self.h_out_w = value if not secondary else self.h_out_s

        mdot_s, d_temps, wf_states, sf_states = self.pinch_calc()

        return abs(self.dt_mean) if self.warning < 100 else 500.0

    def find_pinch(self, secondary=True):
        """
        Vary the secondary fluid enthalpy until the minimum approach temperature is reached.
        Parameters
        ----------
        secondary : bool, optional
            Whether to vary the secondary fluid (True) or the working fluid (False). Default is True.
        Returns
        -------
        float
            The optimized enthalpy of the secondary fluid.
        """
        x0 = copy.copy(self.h_out_s if secondary else self.h_out_w)
        tolerance = 1e-3
    
        try:
            result = minimize(self.pinch_root, x0, args=(secondary,), method='Nelder-Mead', tol=tolerance)
    
            if result.success or result.status == 2:
                if result.status == 2:
                    self.warning = 2
                    self.warning_message = "Minimization problem: " + result.message
                self.h_out_s = result.x[0] if secondary else self.h_out_w
                self.h_out_w = result.x[0] if not secondary else self.h_out_s
                print(f"Optimized enthalpy: {result.x[0]}")  # Debugging-Ausgabe
                return result.x[0]
    
        except Exception as e:
            self._handle_exception(e)
            return 10000
    
        self.warning = 1
        return self.warning


    def pinch_plot(self, plot_fname="", plotting=True):
        """
        Calculate and plot the minimum approach temperature of the heat exchanger.

        Parameters
        ----------
        plot_fname : str, optional
            Filename to save the plot. Default is "".
        plotting : bool, optional
            Whether to plot the results. Default is True.

        Returns
        -------
        m_dot_s : float
            Mass flow rate of the secondary fluid in kg/s.
        d_tempall : np.ndarray
            Temperature differences between the two fluids along the heat exchanger.
        w_array : np.ndarray
            States of the working fluid along the heat exchanger.
        s_array : np.ndarray
            States of the secondary fluid along the heat exchanger.
        """
        m_dot_s, d_tempall, w_array, s_array = self.pinch_calc()

        if plotting:
            self._plot_results(w_array, s_array, plot_fname)

        return m_dot_s, d_tempall, w_array, s_array

    def _calculate_state_array(self, fluid, h_range, points):
        h_array = np.linspace(h_range[0], h_range[1], points)
        values = np.zeros((points, 2))
        values[:, 0] = h_array
        values[:, 1] = fluid.properties.pressure
        return fluid.set_state_v(values, "HP")

    def _check_temperature_consistency(self, d_tempall):
        if np.abs(self.dt_min) - self.d_temp_separation_min < 1e-3:
            self.warning = 900
            self.warning_message = "Below minimum approach temperature!"

        if np.any(d_tempall > 0) and np.any(d_tempall < 0):
            self.warning = 999
            self.dt_mean = 1e6
            self.warning_message = "Temperatures crossing or wrong side!"

    def _plot_heat_exchanger(self, w_array, s_array):
        plot_info = self.plot_info
        print("Plot info:", plot_info)  # Debugging-Ausgabe
        if plot_info["what"][0] == 2:
            data_w = (w_array[:, plot_info["what"][0]] - w_array[:, plot_info["what"][0]].min()) * self.m_dot_w + plot_info["x-shift"][0]
            data_s = (s_array[:, plot_info["what"][0]] - s_array[:, plot_info["what"][0]].min()) * self.m_dot_s + plot_info["x-shift"][1]
            plot_info["ax"].plot(data_w, w_array[:, plot_info["what"][1]], plot_info["col"][0], label=plot_info["label"][0])
            plot_info["ax"].plot(data_s, s_array[:, plot_info["what"][1]], plot_info["col"][1], label=plot_info["label"][1])
        else:
            print(f"H-Ex: plotting only implemented for T-H_dot [2,0]. You requested {plot_info['what']}")

    def _print_verbose(self, d_tempall):
        print(f"Min T-distance {self.dt_min}, Mean T-distance {self.dt_mean}")
        if self.heating > 0:
            print("cond", d_tempall[0], d_tempall[-1], d_tempall.min(), d_tempall.max())
        else:
            print("evap", d_tempall[0], d_tempall[-1], d_tempall.max(), d_tempall.min())

    def _handle_exception(self, e):
        print("find pinch:", type(e))
        print(e.args)
        print(e)
        print("root-exception", self.heating)

    def _plot_results(self, w_array, s_array, plot_fname):
        h_w_plot_array = (w_array[:, 2] - w_array[:, 2].min()) * self.m_dot_w
        fig, ax_one = plt.subplots(1, 1)
        ax_one.plot((s_array[:, 2] - s_array[:, 2].min()) * self.m_dot_s, s_array[:, 0], "v")
        ax_one.plot(h_w_plot_array, w_array[:, 0], "o")
        ax_one.set_xlabel("specific enthalpy flow per mass of secondary fluid / (J / kg)")
        ax_one.set_ylabel("temperature / (K)")
        ax_one.set_title("heat exchanger, simple")
        if plot_fname:
            fig.savefig(plot_fname)



if __name__ == "__main__":
    # two test cases condenser and evaporator:

    FLUID = "Propane * Pentane"  # working fluid
    FLS = "Methanol"  # "Water"  # secondary fluid
    comp = [.50, 0.5]
    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)

    secFlm = cb.fprop.FluidModel(FLS)
    secFluid = cb.fprop.Fluid(secFlm, [1.])
    D_TEMP_MIN = 6.0

    # Condenser, working fluid fixes all, secondary output enthalpy can be varied:
    SEC_TEMP_IN = 300.0
    SEC_TEMP_OUT_MAX = 370.0
    SEC_PRES_IN = 5e5
    H_DOT = 1e3
    state_sec_out = secFluid.set_state([SEC_TEMP_OUT_MAX, SEC_PRES_IN], "TP")

    state_sec_in = secFluid.set_state(
        [SEC_TEMP_IN, SEC_PRES_IN], "TP")  # this is the entering state

    # working fluid

    TEMP_SAT_VAP = SEC_TEMP_OUT_MAX + D_TEMP_MIN
    state_in = myFluid.set_state(
        [TEMP_SAT_VAP, 1.], "TQ")  # find minimum pressure

    WF_TEMP_IN = TEMP_SAT_VAP + D_TEMP_MIN
    WF_TEMP_OUT = SEC_TEMP_IN + D_TEMP_MIN
    state_out = myFluid.set_state([WF_TEMP_OUT, state_in[1]], "TP")

    state_in = myFluid.set_state([myFluid.properties.pressure,
                                  WF_TEMP_IN],
                                 "PT")

    hex0 = StaticHeatExchanger([myFluid, secFluid], H_DOT, state_out[2],
                               h_limit_s=state_sec_out[2],
                               d_temp_separation_min=D_TEMP_MIN)
   
    factor0 = hex0.find_pinch()
    if hex0.warning > 0:
        print(hex0.warning_message)
        
    # now plotting can directly be done in pinch_calc 2024-05-24
    fig_act, ax_act = plt.subplots(1)
    PLOT_INFO ={"fig":fig_act, "ax":ax_act, "what":[2,0],"col":["r:","ko"],
    "label":["work,c","sec,c"], "x-shift":[0,0]}
    hex0.plot_info=PLOT_INFO
    hex0.pinch_calc()
    ax_act.legend()
    #-----------------------------------------------
    #       Previous way
    # ms0, d_tempall0, w0, s0 = hex0.pinch_plot("hex-plot.png")
    

    #  Evaporator: ----------------------------

    SEC_TEMP_IN = 300.0
    SEC_TEMP_OUT = 285
    SEC_PRES_IN = 15e5
    H_DOT = 1e3
    extra = 2
    # D_TEMP_SUPER = 5.
    D_TEMP_MIN = 6.0
    state_sec_out = secFluid.set_state([SEC_TEMP_OUT, SEC_PRES_IN], "TP")
    # this mus be the last set_state before the hex is constructed:
    state_sec_in = secFluid.set_state([SEC_TEMP_IN, SEC_PRES_IN], "TP")

    # WF_TEMP_IN = SEC_TEMP_OUT  # - D_TEMP_MIN
    state_out = myFluid.set_state([SEC_TEMP_IN-D_TEMP_MIN - extra, 1.0], "TQ")
    state_in = myFluid.set_state(
        [SEC_TEMP_OUT-D_TEMP_MIN - extra, state_out[1]], "TP")

    # print("state in", state_in)

    hex1 = StaticHeatExchanger([myFluid, secFluid], H_DOT, state_out[2],
                               h_limit_s=state_sec_out[2],
                               d_temp_separation_min=D_TEMP_MIN)
    # ms1, d_tempall1, w1, s1 = hex1.pinch_calc()
    

    factor_out = hex1.find_pinch()
    if hex1.warning > 2:
        print("Second heat exchanger:", hex1.warning_message)
    else:

        # plotting in the same figure
        PLOT_INFO ={"fig":fig_act, "ax":ax_act, "what":[2,0],"col":["k:","bo"],
        "label":["work,e","sec,e"], "x-shift":[0,0]}
        hex1.plot_info=PLOT_INFO
        hex1.pinch_calc()
        ax_act.legend()



# -*- coding: utf-8 -*-
"""
Class FlowDevice
functions for compressor and expander output state calcultations

so far/here: only for fixed isentropic efficiencies
pumps and throttles added.

((Part of carbatpy.))
Created on Sun May 21 08:51:33 2023

@author: atakan
"""

import carbatpy as cb
import numpy as np


class FlowDevice:
    """Unified themodynamic calculations for single flow devices"""
    def __init__(self, fluid, p_out, m_dot, device_type="machine", **kwargs):
        """
        Calculate the output state and state changes for single flow devices

        Parameters
        ----------
        fluid : Fluid
            The fluid at is entering state.
        p_out : float
            Output pressure in Pa.
        m_dot : float
            mass flow rate in kg/s.
        device_type : TYPE, optional
            'machine' for compressor/expander, 'pump', or 'throttle'. The default is
            "machine".
        **kwargs : dictionary
            Optional parameters for setting the configuration and parameters
            
            - plot info : dictionary,optional, default value is {} contains a Figure, an 
                Axes, a list of What shall be plotted,
                a list with the colour/styles and a list with the labels must be
                passed. in "what", the two numbers coincide with the fluid THERMO
                order. The x-shift can be used in cycle calculations, to shift the
                curves, by the value (it will be added).
                The names in the dictionary are: "fig", "ax", "what","col",
                "label", "x-shift".
            - name : string, optional, the name of this specific device, the default is
                "compressor".
            - calc_type : string, optional, which method shall be used (e.g.: "const_eta",
                "const_h"), default is "const_eta".
            - calc_parameters : dictionary, optional, if further parameters have to be
                given for the calculation. the default is None.
            - verbose : Boolean, optional, the default is False.
            
            
            DESCRIPTION.
            the self.output dictionary contains the exit state values, and work and power.
            The self.fluid is at the exit state. Check self.warning and self.warning_message.
            There is also a self.print_device function for inspection. If a figure is passed
            by plot_info the plot will be updated.

        Returns
        -------
        None.

        """
        self.fluid = fluid
        self.p_out = p_out
        self.m_dot = m_dot
        self.device_type = device_type
        self.plot_info = kwargs.get("plot_info", {})
        self.name = kwargs.get("name", "compressor")
        self.calc_type = kwargs.get("calc_type", "const_eta")
        self.calc_parameters = kwargs.get("calc_parameters", None)
        self.verbose = kwargs.get("verbose", False)
        self.warning = 0
        self.warning_message = "All o.k."
        self.input = fluid.properties.state
        self.output = None

        match device_type:  # noeq
            case "machine":
                self.compressor()
            case "pump":
                self.pump()
            case "throttle":
                self.throttle()
            case _:
                print(f"Not implemented device: {device_type}")

    def compressor(self):
        """
        compressor or expander output state calculation

        so far only for a constant isentropic efficiency, according to the pressure
        change an expansion or compression is detected and handled.

        Parameters
        ----------

        p_out : float
            output pressure.
        eta_s : float
            isentropic efficiency.
        fluid : fprop.Fluid
            entering fluid, including properties, composition, and model.
        m_dot : float, optional
            mass flow rate (in kg/s). Default is 1
        calc_type : string, optional
            how to calculate, so far, only one implemented. The default is
            "const_eta".
        name : string, optional
            name of the device. The default is "compressor".
        plot_info : dictionary, optional
            if not empty a Figure, an Axes, a list of What shall be plotted,
            a list with the colour/styles and a list with the labels must be
            passed. in "what", the two numbers coincide with the fluid THERMO
            order. The x-shift can be used in cycle calculations, to shift the
            curves, by the value (it will be added).
            The names in the dictionary are: "fig", "ax", "what","col",
            "label", "x-shift". Default is empty.

        Returns
        -------
        state_out : array of float
            compressor output state containing [T,p,h,v,s,q].
        work_specific : float
            work per kg of fluid, positive for compressor; units:J/kg.

        """
        state_in = self.fluid.properties.state
        expander = False
        if self.fluid.properties.pressure > self.p_out:
            expander = True

        if self.calc_type == "const_eta":
            if self.calc_parameters["eta_s"]:
                eta_s = self.calc_parameters["eta_s"]
            else:
                self.warning = 99
                self.warning_message = f"No 'eta_s' provided, a value of 1 will be used!"
                self.calc_parameters["eta_s"] = eta_s = 1.0

            self.fluid.set_state(
                [self.fluid.properties.entropy, self.p_out], "SP")

            diff_enthalpy_s = self.fluid.properties.enthalpy-state_in[2]

            if expander:
                work_specific = diff_enthalpy_s * eta_s
            else:
                work_specific = diff_enthalpy_s / eta_s

            state_out = self.fluid.set_state(
                [state_in[2] + work_specific, self.p_out], "HP")

        else:
            self.warning = 100
            self.warning_message = f""""The option{calc_type} is not yet implemented for
            compressors"""
            raise Exception(
                self.warning_message)

        power = self.m_dot * work_specific

        #
        self.output = {"state_out": state_out,
                       "work_specific": work_specific,
                       "power": power}
        self.plot_temp_h_flow()

    def pump(self):
        """
        Calculate the exit state of a pump assuming an incompressible fluid.

        Only formulated for constant isentropic efficiency

        Parameters
        ----------
        p_out : float
            output pressure.
        eta_s : float
            isentropic efficiency.
        fluid : fprop.Fluid
            entering fluid, including properties, composition, and model.
        m_dot : float, optional
            mass flow rate (in kg/s). Default is 1
        calc_type : string, optional
            how to calculate, so far, only one implemented. The default is
            "const_eta".
        name : string, optional
            name of the device. The default is "pump".
        plot_info : dictionary, optional
            if not empty a Figure, an Axes, a list of What shall be plotted,
            a list with the colour/styles and a list with the labels must be
            passed. in "what", the two numbers coincide with the fluid THERMO
            order. The x-shift can be used in cycle calculations, to shift the
            curves, by the value (it will be added).
            The names in the dictionary are: "fig", "ax", "what","col",
            "label", "x-shift". Default is empty.

        Returns
        -------
        state_out : array of float
            compressor output state containing [T,p,h,v,s,q].
        work_specific : float
            work per kg of fluid, positive for compressor; units:J/kg.

        """
        state_in = self.input
        if self.calc_type == "const_eta":

            if self.calc_parameters["eta_s"]:
                eta_s = self.calc_parameters["eta_s"]
            else:
                self.warning = 99
                self.warning_message = f"No 'eta_s' provided, a value of 1 will be used!"
                self.calc_parameters["eta_s"] = eta_s = 1.0

            work_is = state_in[3] * (self.p_out - state_in[1])
            if work_is > 0:
                work_specific = work_is / eta_s
            else:
                work_specific = work_is * eta_s
            h_out = state_in[2] + work_specific
            state_out = self.fluid.set_state([h_out, self.p_out], "HP")
        else:
            self.warning = 130
            self.warning_message = f"The option{calc_type} is not yet implemented for pumps"
            raise Exception(self.warning_message)
        power = self.m_dot * work_specific

        self.output = {"state_out": state_out,
                       "work_specific": work_specific,
                       "power": power}
        self.plot_temp_h_flow()

    def throttle(self):
        """
        throttle output state calculation

        so far only for a constant enthalpy

        Parameters
        ----------
        p_out : float
            output pressure.
        fluid : fprop.Fluid
            entering fluid, including properties, composition, and model.
        m_dot : float, optional
            mass flow rate (in kg/s). Default is 1
        calc_type : string, optional
            how to calculate, so far, only one implemented. The default is
            "const_h".
        name : string, optional
            name of the device. The default is "throttle".
        plot_info : dictionary, optional
            if not empty a Figure, an Axes, a list of What shall be plotted,
            a list with the colour/styles and a list with the labels must be
            passed. in "what", the two numbers coincide with the fluid THERMO
            order. The x-shift can be used in cycle calculations, to shift the
            curves, by the value (it will be added).
            The names in the dictionary are: "fig", "ax", "what","col",
            "label", "x-shift". Default is empty.

        Returns
        -------
        state_out : array of float
            compressor output state containing [T,p,h,v,s,q].

        """
        state_in = self.fluid.properties.state
        if self.calc_type == "const_h":
            state_out = self.fluid.set_state([self.fluid.properties.enthalpy,
                                              self.p_out], "HP")
        else:
            raise Exception(
                f"The option{self.calc_type} is not yet implemented for throttles")

        self.output = {"state_out": state_out,
                       "work_specific": 0.0,
                       "power": 0.0}
        self.plot_temp_h_flow()

        #
        return state_out

    def plot_temp_h_flow(self):
        """
        plotting a T-H-dot diagram for simple flows (compressor, throttle etc.)

        Parameters
        ----------
        _state_in : np.array
            entering state [T,p,h,v,s,...].
        _state_out : np.array
            exiting state.
        _m_dot : float
            mass flow rate (kg/s).
        _plot_info : dictionary
            if not empty a Figure, an Axes, a list of What shall be plotted,
            a list with the colour/styles and a list with the labels must be
            passed. in "what", the two numbers coincide with the fluid THERMO
            order. The x-shift can be used in cycle calculations, to shift the
            curves, by the value (it will be added).
            The names in the dictionary are: "fig", "ax", "what","col",
            "label", "x-shift".

        Returns
        -------
        None.

        """
        if len(self.plot_info) > 0:
            if self.plot_info["what"][0] == 2:
                data = np.array([self.input[self.plot_info["what"][0]],
                                 self.output["state_out"][self.plot_info["what"][0]]]) * self.m_dot \
                    + self.plot_info["x-shift"][0]
                self.plot_info["ax"].plot(data,
                                          [self.input[self.plot_info["what"][1]],
                                           self.output["state_out"][self.plot_info["what"][1]]],
                                          self.plot_info["col"][0],
                                          label=self.plot_info["label"][0])
            else:
                self.warning = 130
                self.warning_message = f"Pump: plotting only implemented fot T-H_dot [2,0]. You requested{_plot_info['what']}"
                print(self.warning_message)

    def print_device(self):

        print(f"\nDevice: {self.device_type}, {self.calc_type}")
        print(f"Parameters, {self.calc_parameters}")
        print(f"Input, {self.input}")
        print(f"Output: {self.output}")
        if self.warning:
            print(f"Device: {self.warning}, {self.warning_message}")


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    FLUID = "Propane * Pentane"
    comp = [.80, 0.2]
    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)
    P_LOW = 1e5
    T_IN = 310.
    DT_IN_LIQ = -5
    state_in_act = myFluid.set_state([T_IN, P_LOW], "TP")
    P_OUT = 10e5
    ETA_S = .7
    M_DOT = 1e-3
    fig0, ax0 = plt.subplots()
    PLOT_INFO = {"fig": fig0, "ax": ax0, "what": [2, 0], "col": ["r:", "k"],
                 "label": ["compressor", "xx"], "x-shift": [0, 0]}

    # Compressor-------------
    compressor = FlowDevice(myFluid, P_OUT, M_DOT,
                            device_type="machine",
                            name="compressor",
                            calc_type="const_eta",
                            calc_parameters={"eta_s": ETA_S},
                            plot_info=PLOT_INFO)
    # state_o, work, power_c = compressor(myFluid, P_OUT, M_DOT, ETA_S, device_type="machine",
    # plot_info=PLOT_INFO)
    print(myFluid.properties.temperature, compressor.output)
    # print("\nCompressor", state_in_act, "\n", state_o, "\n", state_o-state_in_act)
    PLOT_INFO["col"] = ["k", ""]
    PLOT_INFO["label"] = ["expander", ""]

    expander = FlowDevice(myFluid, P_LOW, M_DOT,
                          device_type="machine",
                          name="expander",
                          calc_type="const_eta",
                          calc_parameters={"eta_s": ETA_S},
                          plot_info=PLOT_INFO)
    print("\nExpander:", expander.output)

    # Pump, incompressible:

    state_in_p = myFluid.set_state([P_LOW, 0], "PQ")
    state_in_p = myFluid.set_state([P_LOW, state_in_p[0]+DT_IN_LIQ], "PT")
    PLOT_INFO["col"] = ["bv:", ""]
    PLOT_INFO["label"] = ["pump", ""]
    pump = FlowDevice(myFluid, P_OUT, M_DOT,
                      device_type="pump",
                      name="pump-A",
                      calc_type="const_eta",
                      calc_parameters={"eta_s": ETA_S},
                      plot_info=PLOT_INFO)
    print(
        f"pump work output: {pump.output['work_specific']:.3f} J/kg, \npump all:{pump.output}")
    
    # throttle:
    PLOT_INFO["col"] = ["go:", ""]
    PLOT_INFO["label"] = ["throttle", ""]
    state_in_p = myFluid.set_state([P_OUT, state_in_p[0]+DT_IN_LIQ], "PT")
    throttle = FlowDevice(myFluid, P_OUT, M_DOT,
                          device_type="throttle",
                          name="throttle-A",
                          calc_type="const_h",
                          calc_parameters={},
                          plot_info=PLOT_INFO)
    throttle.print_device()
    
    # if you want a legend etc
    ax0.legend()
    ax0.set_ylabel("T/K")
    ax0.set_xlabel("$\dot h$/ kW")



# -*- coding: utf-8 -*-
"""
Some constants for carbatpy usage are set here

* The two (!) Refprop installation directories (with dll, fluids etc.)
* he directory, where the results shall be written.

Created on Sun Nov  5 16:01:02 2023

@author: atakan
"""

import os
import copy
# global _T_SURROUNDING, _P_SURROUNDING, _RESULTS_DIR, _CARBATPY_BASE_DIR
_T_SURROUNDING = 288.15 # K
_P_SURROUNDING =1.013e5  # Pa
TREND = {"TREND_INSTALLED":True,
          "USE_TREND":False, 
          "TREND_DLL":"", 
          'TREND_PATH':""}

directory = os.getcwd()
try:
    _CARBATPY_BASE_DIR = os.environ["CARBATPY_BASE_DIR"]
except:
    _CARBATPY_BASE_DIR = directory

# The two installations of REFPROP , one for the working fluid
# and one for the secondary fluid. With only one installation, the instances
# mix up ...
os.environ['RPPREFIX'] = r'C:/Program Files (x86)/REFPROP'
os.environ['RPPREFIXs'] = r'C:/Program Files (x86)/REFPROP/secondCopyREFPROP'
if TREND["TREND_INSTALLED"]:
    try:
        TREND["TREND_DLL"] = os.environ['TREND_DLL']
        TREND["TREND_PATH"] = os.environ['TREND_PATH']
    except:
        print("Trend not found! Check the environmentvariable TREND_DLL, TREND_PATH")
        TREND_DLL = ""
        TREND_PATH = ''
        TREND["TREND_INSTALLED"] = False
    
try:
    _RESULTS_DIR = os.environ['CARBATPY_RES_DIR']
except:
    try:
        _RESULTS_DIR = os.environ['TEMP']
    except:
        try:
            
            _RESULTS_DIR = directory + r"\\tests\\test_files"
        except Exception as no_file:
            print("Please set the envirionment variable: CARBATPY_RES_DIR !", no_file)
            
# Default values:-------------------------------------------------
fl_properties_names = ("Temperature", "Pressure", "spec_Enthalpy",
                        "spec_Volume", "spec_Entropy", "quality",
                        "spec_internal_Energy",
                        "viscosity", "thermal_conductivity",
                        "Prandtl_number", "k_viscosity", "molecular_mass",
                        "speed_of_sound")
THERMO_STRING = "T;P;H;V;S;QMASS;E"
TRANS_STRING = THERMO_STRING + ";VIS;TCX;PRANDTL;KV;M;W"
THERMO_TREND = request = ["T", "P", "H", "D",
                           "S", "QEOS", "U"]  # careful density not volume
TRANS_STRING = THERMO_STRING + ";VIS;TCX;PRANDTL;KV;M;W"
TRANS_TREND = copy.copy(THERMO_TREND)
TRANS_TREND.extend(["ETA", "TCX", "CP"])

CB_FLUID_DEFAULT = {"PROPS": "REFPROP",  # choice: "TREND" or "REFPROP"
                    "DLL_SELECT": "2dll",  # choice: "2dll" or "dll" 
                    "UNITS": "MASS BASE SI",
                    "Property_Names":fl_properties_names,
                    "THERMO_STRING":THERMO_STRING,
                    "TRANS_STRING":TRANS_STRING ,
                    "THERMO_TREND": THERMO_TREND,
                    "TRANS_TREND": TRANS_TREND,
                    "TREND":TREND,
                    "RPPREFIX":os.environ['RPPREFIX'] 
                    
                    
                    
    }
# This dictionary will be generally imported with carbatpy
CB_DEFAULTS={
    "Fluid_Defaults": CB_FLUID_DEFAULT,
    "General":{"T_SUR":_T_SURROUNDING,
               "P_SUR": _P_SURROUNDING,
               "RES_DIR": _RESULTS_DIR
               }
    }